# (c) 2021-05-09 Osamu NAKAMURA
#
# This script performs the followings.
#
#   1. Estimate Confidence interval of Krippendorff's alpha
#   2. Inter-coder test


####################
#################### Initial Settings
####################

library("irr")
#library("readxl")
#library("stats")
  # "Stats" is read by default. So, it is probably unnecessary to read them here.
  # readxl is to read Excel files in future ver., but currently not yet
  #   supported.

n_boot   <- 1000
  # Number of bootstrap loops. About 1000 times.
  # - If you think the result is not accurate enough, increase n_boot.
  # - If you think it takes too much time without improving accuracy, decrease
  #   n_boot.

  # note.
  # If you specify set.seed(), you can fix the seed of a random number,
  # and you can obtain always the same result for the same data.
  # For this script, I don't think this treatment is necessary considering the
  #   purpose.

v_type   <- "nominal"
#v_type   <- "ordinal"
#v_type   <- "interval"
  # Type of your coding variable. See irr's help for details.
  #   nominal   (Japanese: "meigi syakudo")
  #   ordinal   (Japanese: "junjyo syakudo")
  #   interval  (Japanese: "tou-kankaku syakudo")
  #   ratio     (Japanese: "hiritsu syakudo")

c_lvl    <- c(0.005, 0.025, 0.050, 0.500, 0.950, 0.975, 0.995)
  # Specify in c_lvl the %-ile you want to output when the values obtained
  # by bootstrap are sorted in order.


####################
#################### Preparation for Data Loading
####################

arg_inpt         <- commandArgs(trailingOnly = T)
  # Read the arguments on the command line when executing this script.


arg_dflt         <- c("data_test.csv")
#arg_dflt         <- c("data_test_HT_20231214.csv")
#arg_dflt         <- c("data_test_KR_2coders.csv")
#arg_dflt         <- c("data_test_KR_2coders_factor.csv")
#arg_dflt         <- c("data_test_KR_3coders.csv")
#arg_dflt         <- c("data_test_KR_4coders.csv")
#arg_dflt         <- c("data_test_KR_4coders_factor.csv")
#arg_dflt         <- c("data_test_KR_4coders_factor_JP.csv")
#arg_dflt         <- c("data_test_KR_5coders.csv")
arg_na           <- is.na(arg_inpt[1:1])
  # arg_dflt: default argument values
  #           Set the default values for arguments so that "source" on R
  #           command-line would work without arguments even when Windows batch
  #           is not used.
  # arg_na:   Element number of the argument that was N/A when this script is
  #           executed.
arg_inpt[arg_na] <- arg_dflt[arg_na]
  # Assign default values for arguments that are N/A.

fl_inp           <- arg_inpt[1]
  # Determine data file name (first argument is the data file name)


####################
#################### Data Loading
####################

##########
########## Case 1: When the recorded variable is an interval/ordinal/ratio scale
##########         (i.e., numerical data)
##########

if ( v_type == "ordinal" | v_type == "interval" | v_type == "ratio"){

  x_tmp1     <- read.table(fl_inp, sep=",", na.strings="")
  #x_tmp1     <- read.table(fl_inp, sep=",", na.strings="", fileEncoding="cp932")
  #x_tmp1     <- read.table(fl_inp, sep=",", na.strings="", fileEncoding="UTF-8")
    # Read coder coding results from the data file.
    # blanks treated as NA.

  d_cdr1     <- data.matrix(x_tmp1)
  d_cdr      <- d_cdr1
    # Keep d_cdr1 for de-bugging purpose.
 
    # note.
    # This conversion(data.matrix) can not apply to data files recorded
    # in a form of factors like "Positive" and "Negative".


##########
########## Case 2: When the recorded variable is a nominal scale (including factor)
########## Convert factor to numeric (exclude missing values(NA))
##########

} else if ( v_type == "nominal" ){

  # The following algorithm does not work if the data file is recorded
  # in a form of ordinal/interval/ratio, because original "1 2 5 5 6"
  # could be converted to "1 2 4 4 3", for example.


  x_tmp2     <- read.csv(fl_inp, header=FALSE, na.strings="")
  #x_tmp2     <- read.csv(fl_inp, header=FALSE, na.strings="", fileEncoding="cp932")
  #x_tmp2     <- read.csv(fl_inp, header=FALSE, na.strings="", fileEncoding="UTF-8")
    # Read coder coding results from the data file.
    # blanks treated as NA.

  x_tmp2_num <- x_tmp2
  lev_f      <- unique(na.omit(unlist(x_tmp2_num)))
    # lev_f: list of levels in x_tmp2 as factor, excluding NA.

  for (col in 1:ncol(x_tmp2_num)){
    col_val                 <- x_tmp2_num[, col]
    col_val[is.na(col_val)] <- NA
    x_tmp2_num[, col]       <- match(col_val, lev_f)
  }
    # While excluding NA, factor is converted to an integer by "match"
    # and assigned back to x_tmp_num.

  d_cdr2     <- as.matrix(x_tmp2_num)
  d_cdr      <- d_cdr2
    # Convert to matrix.
    # Keep d_cdr2 for de-bugging purpose.


##########
########## Case 3: Error in v_type
########## 

} else {
  print("Error in v_type.")
  stop()
}


####################
#################### Calculation of Krippendorff's alpha
####################

a_tmp    <- kripp.alpha(d_cdr, v_type)
a_org    <- a_tmp$value
  # Simple calculation of alpha.
  # According to irr's help, "$value" contains the alpha value.


##########
########## Output
##########

cat(sprintf("\n"))
cat(sprintf("Krippendorff's alpha = %f\n", a_org))
  # Output the correct alpha value of the data first.


##########
########## Debug
##########

#a_tmp1_n   <- kripp.alpha(d_cdr1, "nominal" )
#a_tmp1_o   <- kripp.alpha(d_cdr1, "interval")
#a_tmp2_n   <- kripp.alpha(d_cdr2, "nominal" )
#a_tmp2_o   <- kripp.alpha(d_cdr2, "interval")
#a_org1_n   <- a_tmp1_n$value
#a_org1_o   <- a_tmp1_o$value
#a_org2_n   <- a_tmp2_n$value
#a_org2_o   <- a_tmp2_o$value
#cat(sprintf("Krippendorff's alpha data1 nominal = %f\n", a_org1_n))
#cat(sprintf("Krippendorff's alpha data1 ordinal = %f\n", a_org1_o))
#cat(sprintf("Krippendorff's alpha data2 nominal = %f\n", a_org2_n))
#cat(sprintf("Krippendorff's alpha data2 ordinal = %f\n", a_org2_o))
#stop()



####################
#################### Error estimation of alpha by bootstrap
####################

# The following web site is referred to for how to perform bootstrap:
#   Hiroshi TAROMARU "Introduction to Bootstrap" (2016-04-03)
#   http://tarohmaru.web.fc2.com/R/bootstrapping.html

n_obj    <- length(d_cdr[1,])
n_cdr    <- length(d_cdr[,1])
      # n_obj: Number of coding objects (e.g., number of articles).
      # n_cdr: Number of coders.

a_boot   <- rep(NA, n_boot)
      # First prepare an empty vector "a_boot" to record alpha for n_boot
      # loop from now on.


for(i in 1:n_boot){
      # loop for bootstrap

  smplid    <- sample(1:n_obj, n_obj, replace=TRUE)
      # First determine the element ID to be extracted by bootstrap,
      # and save their IDs in smplid.
      #  - replace=TRUE: sampling with replacement in bootstrap
      #    (="fukugen cyu-syutsu" = sampling with duplicates allowed).

      # note:
      #   I checked the algorithm be working well by doing the followings:
      #   smplid <- 1:n_obj (That is, the original data is copied to smplid)
      #   n_boot <- 1       (That is, bootstrap is not performed.)

  c_smpl    <- d_cdr[, smplid]
      # Extract coder data for bootstrap(smplid) and assign them to c_smpl.
      # c_smpl is a pseudo data of "d_cdr" by bootstrap.

  a_tmp     <- kripp.alpha(c_smpl, v_type)
      # Calculate kripp.alpha by the pseudo data.
  a_boot[i] <- a_tmp$value
      # Among the results, record alpha value to a_boot
      # (alpha is stored in $value, according to irr help).

}


##########
########## Output-1
########## Percentiles are taken directly from the distribution.
##########

# note. define c_lvl prior to this section.


cat(sprintf("\n"))
cat(sprintf("Bootstrap distribution:("))
cat(c_lvl)
cat(sprintf(")\n"))

#cat(quantile(a_boot, c(0.025, 0.050, 0.500, 0.950, 0.975)), "\n")
 cat(quantile(a_boot, c_lvl                               ), "\n")
      # Output each percentile as a result of bootstrap.

      # If you want to write out the results to a file etc.,
      # continue here.


##########
########## Output-2
########## Standard Error is calculated to compare with Output-1.
##########

# Note.
# The consistency of Output-1 and Output-2 has already been confirmed.
# Therefore, the following sections are commented out.

# # In this case the Standard Error is simply equal to Standard deviation
# # of bootstrap trials. Refer to texts on SE of bootstrap.
# 
# tmp_av <- mean(    a_boot )
# tmp_se <- sqrt(var(a_boot))
#   # - tmp_av: Mean of Bootstrap trials.
#   # - tmp_se: Standard Error of Bootstrap (= std-dev of trial values).
# 
# cat(sprintf("Mean=          %f\n", tmp_av))
# cat(sprintf("StandardError= %f\n", tmp_se))
# cat(sprintf("95percent-confident interval: %f -- %f\n", tmp_av-1.96*tmp_se, tmp_av+1.96*tmp_se))
# cat(sprintf("90percent-confident interval: %f -- %f\n", tmp_av-1.65*tmp_se, tmp_av+1.65*tmp_se))
#   # Use "percent" instead of "%" in sprintf.



####################
#################### Inter-Coder Test
####################

# This section is to check what alpha value should be obtained with 1-to-1
# coder pair each. This section is added to check if there are some coders
# who achieves low match with anyone else.

cat(sprintf("\n"))
cat(sprintf("Inter-coder test\n"))

a_avr     <- rep(NA, n_cdr)
        # First prepare an empty vector a_avr for coder-i to record the average
        # value of alpha with others.

for(i in 1:n_cdr){
        # loop for each coder.

  a_tmp   <- rep(NA, n_cdr)
        # First prepare an empty vector a_tmp for coder-i to record alpha
        # with coder-j in pair.

  for(j in 1:n_cdr){
        # Loop to make a pair of coder-i and -j
        # (this script tries all combinations).

    if( i != j ){

      c_pair   <- d_cdr[c(i,j),]
        # Extract the data of pair of coder-i and -j.

      a_pair   <- kripp.alpha(c_pair, v_type)
      a_tmp[j] <- a_pair$value
        # Calculate the alpha of that pair and record the result.

      a_smp    <- sum(c_pair[1,]==c_pair[2,], na.rm=TRUE)/length(c_pair[1,])
        # a_smp is the simle agreement between the two coders.
        # - "sum(x==y)"" counts the number of elements with x==y.
        # - "na.rm=TRUE" option excludes elements with NA.
        # - length(x) is the number of elements in x.

      cat(sprintf("Coder %03d vs. Coder %03d Kripp-alpha= %f  Simple-agreement= %f\n", i,j, a_tmp[j], a_smp))
        # I decided not to save all the result.
        # Instead, the detailed result is output here.

    }
  }

  a_avr[i] <- mean(a_tmp, na.rm=T)
        # Record in a_avr the average of alpha of coder-i with 1-to-1 other
        #   coders. Remove when i=j by "na.rm=T" because the value for i=j is
        #   left N/A.

}


##########
########## Output Summary
##########

cat(sprintf("\n"))

a_rnk <- rank(-a_avr, ties.method="min")
        # Calculate the rank of average alpha when each coder is paired with
        # other coders. a_avr is given a minus sign because the alpha is in
        # descending order.

for(i in 1:n_cdr){

  cat(sprintf("Coder %03d rank= %3d average= %f\n", i, a_rnk[i], a_avr[i]))

}


####################
#################### Distribution of alpha in the case of random coding
####################

# I would like to see how random the alpha of random coding actually is.

##########
########## prepare 1-D array with coding data placed on a single line
##########

d_cd1    <- d_cdr[1,]
        # d_cd1: Same as d_cdr, but data is placed on a single line.
        # - e.g. 3 rows (coders) x 10 columns (objects) is transformed
        #        to an 1-dimensional array with 30 data elements.
        # - This is used for random sampling.
        # - Here data of 1st coder is copied to d_cd1.

for(i in 2:n_cdr){
  d_cd1 <- append(d_cd1, d_cdr[i,])
}
        # Data of subsequent coders are copied to d_cd1, too.


##########
########## simulation of random coding
##########

a_b_rd   <- rep(NA, n_boot)
        # First prepare an empty vector "a_b_rd" to record alpha of random
        # coding for n_boot loop.

for(i in 1:n_boot){
        # loop for bootstrap
        # each loop tries simulation of random coding.

  d_rnd <- d_cdr
        # copy the original data to create a space to store simulation data.
        # d_rnd: coding data when random coding is simulated.


  for (j in 1:n_obj){
        # loop to simulate random coding for n_obj objects.

    d_tmp <- sample(d_cd1, n_cdr, replace=FALSE)
        # perform random sampling for coders.
        # - randomly extract n_cdr elements from d_cd1 and assign to d_tmp.
        # - replace=FALSE: sampling without replacement
        #   (i.e., sampling without duplicates).

    d_rnd[,j] <- d_tmp
        # assign random coding to d_rnd[,j] as data for j-th object.
  }

  a_tmp     <- kripp.alpha(d_rnd, v_type)
        # Calculate kripp.alpha by the pseudo data of random coding.
  a_b_rd[i] <- a_tmp$value
        # Among the results, record alpha value to a_b_rd
        # (alpha is stored in $value, according to irr help).

}

##########
########## Output
##########

cat(sprintf("\n"))
cat(sprintf("Alpha distribution of random coding:("))
cat(c_lvl)
cat(sprintf(")\n"))

#cat(quantile(a_b_rd, c(0.025, 0.050, 0.500, 0.950, 0.975)), "\n")
 cat(quantile(a_b_rd, c_lvl                               ), "\n")
      # Output each percentile as a result of bootstrap.

